function [ U1, U2, P1, P2, V ] = solveIISPH( X1, X2, A12, D12, k, alphas, beta, mu, lambda, gamma, bits)

% Reference:
% Z. Chen, F. Zhong, G. Min, Y. Leng, and Y. Ying, 
% “Supervised Intra- and Inter-Modality Similarity Preserving Hashing for Cross-Modal Retrieval,?
% IEEE Access, vol. 6, pp. 27796?7808, 2018.
%
% Written by Fangming Zhong
%
%  X1: d1*n
%  X2: d2*n
%  U1: d1*bits
%  U2: d2*bits
%  V:  bits*n
%  P1: bits*d1
%  P2: bits*d2
%
%% random initialization

[d1, n] = size(X1);
[d2, ~] = size(X2);
V = rand(bits, n);
U1 = rand(d1,bits);
U2 = rand(d2,bits);
P1 = eye(bits, d1);
P2 = eye(bits, d2);
threshold = 0.01;
lastF = 99999999999;
iter = 1;

A21 = A12';
D21 = D12';

%compute L1 and L2, based P1*X1, P2*X2, and k
DIST1 = calc_distance((P1*X1)',(P1*X1)','L2');
DIST2 = calc_distance((P2*X2)',(P2*X2)','L2');

%add label information to dist matrix
Diss = A12;
for i=1:n
    Diss(i,i) = 0;
end

beta2 = mean(DIST1(:));
beta3 = mean(DIST2(:));

Dis = Diss;
Dis = Dis .* exp(-DIST1/(0.01*beta2));
Dis(Dis==0) = 1;
DIST1 = DIST1 .* Dis;

Dis = Diss;
Dis = Dis .* exp(-DIST2/(0.01*beta3));
Dis(Dis==0) = 1;
DIST2 = DIST2 .* Dis;

[L1, ~, ~] = calc_Laplacian(DIST1,k);
[L2, ~, ~] = calc_Laplacian(DIST2,k);

    
%% compute iteratively
while (true)
	% update U1 and U2
    U1 =  X1 * V' / (V*V' + gamma / alphas(1) * eye(bits));   
    U2 =  X2 * V' / (V*V' + gamma / alphas(2) * eye(bits));
    
	% update V
    V = ( alphas(1) * U1' * U1 + alphas(2) * U2' * U2 + beta * eye(bits) + beta * eye(bits) + gamma * eye(bits) ) \ ( alphas(1) * U1' * X1 + alphas(2) * U2' * X2 + beta * P1 * X1 + beta * P2 * X2);

    %update P1 and P2
    P1 = (beta*V*X1' + mu*P2*X2*A21*X1') / (beta*X1*X1' + mu*X1*D12*X1' + lambda*X1*L1*X1' + gamma * eye(d1));
    P2 = (beta*V*X2' + mu*P1*X1*A12*X2') / (beta*X2*X2' + mu*X2*D21*X2' + lambda*X2*L2*X2' + gamma * eye(d2));
    
    % compute new L1, L2 again
    DIST1 = calc_distance((P1*X1)',(P1*X1)','L2');
    DIST2 = calc_distance((P2*X2)',(P2*X2)','L2');
    
    
    % add supervised information
    %
    beta2 = mean(DIST1(:));
    beta3 = mean(DIST2(:));

    Dis = Diss;
    Dis = Dis .* exp(-DIST1/(0.1*beta2));
    Dis(Dis==0) = 1;
    DIST1 = DIST1 .* Dis;

    Dis = Diss;
    Dis = Dis .* exp(-DIST2/(0.1*beta3));
    Dis(Dis==0) = 1;
    DIST2 = DIST2 .* Dis;

    
    [L1, ~, ~] = calc_Laplacian(DIST1,k);
    [L2, ~, ~] = calc_Laplacian(DIST2,k);
    
        
    % compute objective function
    norm1 = norm(X1 - U1 * V, 'fro') ^ 2; 
    norm2 = norm(X2 - U2 * V, 'fro') ^ 2;
    norm3 = norm(V - P1 * X1, 'fro') ^ 2 + norm(V - P2 * X2, 'fro') ^ 2;
    norm4 = trace(P1 * X1 * D12 * X1' * P1' + P2 * X2 * D21 * X2' * P2') - trace(P1 * X1 * A12 * X2' * P2' + P2 * X2 * A21 * X1' * P1');
    norm5 = trace(P1 * X1 * L1 * X1' * P1' + P2 * X2 * L2 * X2' * P2');
    norm6 = gamma * ( norm(U1, 'fro') ^ 2 + norm(U2, 'fro') ^ 2 + norm(P1, 'fro') ^ 2 + norm(P2, 'fro') ^ 2 + norm(V, 'fro') ^ 2);
    
    
    currentF= alphas(1) * norm1 + alphas(2) * norm2 + beta * norm3 + mu * norm4 + lambda* norm5 + gamma * norm6;
    
    fprintf('\nobj at iteration %d: %.4f\n reconstruction error for collective matrix factorization: %.4f,\n reconstruction error for linear projection: %.4f,\n inter-modality error: %.4f,\n intra-modality error : %.4f,\n regularization term: %.4f\n\n', ...
        iter, currentF, alphas(1) * norm1 + alphas(2) * norm2, norm3, norm4, norm5, norm6);
    if (lastF - currentF) < threshold
        fprintf('algorithm converges...\n');
        fprintf('final obj: %.4f\n reconstruction error for collective matrix factorization: %.4f,\n reconstruction error for linear projection: %.4f,\n inter-modality error: %.4f,\n intra-modality error : %.4f,\n regularization term: %.4f\n\n', ...
            currentF, alphas(1) * norm1 + alphas(2) * norm2, norm3, norm4, norm5, norm6);
        return;
    end
    iter = iter + 1;
    lastF = currentF;
    
    if iter > 1
       break;
    end
    
end
end

